@@ -1,5 +1,6 @@
 #include "gcin.h"
 #include "gtab.h"
+#include <signal.h>
 #include <X11/extensions/XTest.h>
 
 
@@ -10,103 +11,144 @@ extern Display *dpy;
 extern XIMS current_ims;
 static IMForwardEventStruct *current_forward_eve;
 #endif
+extern gboolean win_kbm_inited;
 
-static char callback_str_buffer[32];
+static char *callback_str_buffer;
 Window focus_win;
 static int timeout_handle;
 
-static void send_fake_key_eve()
+void send_fake_key_eve(KeySym key)
 {
-  KeyCode kc_shift_l = XKeysymToKeycode(dpy, XK_Shift_L);
-  XTestFakeKeyEvent(dpy, kc_shift_l, True, CurrentTime);
-  XTestFakeKeyEvent(dpy, kc_shift_l, False, CurrentTime);
+  KeyCode kc = XKeysymToKeycode(dpy, key);
+  XTestFakeKeyEvent(dpy, kc, True, CurrentTime);
+  XTestFakeKeyEvent(dpy, kc, False, CurrentTime);
 }
 
+
 void send_text_call_back(char *text)
 {
+  callback_str_buffer = realloc(callback_str_buffer, strlen(text));
   strcpy(callback_str_buffer, text);
-  send_fake_key_eve();
+  send_fake_key_eve(XK_Shift_L);
 }
 
 
 ClientState *current_CS;
-ClientState temp_CS;
+static ClientState temp_CS;
 
 gboolean init_in_method(int in_no);
 
-char *output_buffer, *output_buffer_bak;
-int  output_bufferN;
+char *output_buffer;
+int output_bufferN;
+static char *output_buffer_raw, *output_buffer_raw_bak;
+static int output_buffer_rawN;
 
 void clear_output_buffer()
 {
-  char *t;
-
-  t = output_buffer;
-  output_buffer = output_buffer_bak;
-  output_buffer_bak = t;
-
   if (output_buffer)
     output_buffer[0] = 0;
-
   output_bufferN = 0;
+
+  char *t = output_buffer_raw;
+  output_buffer_raw = output_buffer_raw_bak;
+  output_buffer_raw_bak = t;
+
+  if (output_buffer_raw)
+    output_buffer_raw[0] = 0;
+  output_buffer_rawN = 0;
 }
 
 gboolean gb_output = FALSE;
 
+void toggle_gb_output()
+{
+  gb_output = !gb_output;
+}
+
+static void append_str(char **buf, int *bufN, char *text, int len)
+{
+  int requiredN = len + 1 + *bufN;
+  *buf = realloc(*buf, requiredN);
+  (*buf)[*bufN] = 0;
+  strcat(*buf, text);
+  *bufN += len;
+}
+
 void send_text(char *text)
 {
+  char *filter;
+
   if (!text)
     return;
   int len = strlen(text);
 
+  append_str(&output_buffer_raw, &output_buffer_rawN, text, len);
+
   char *utf8_gbtext = NULL;
 
   if (gb_output) {
-    u_int rn = 0 , wn = 0;
-    GError *err = NULL;
-    char *big5 = g_convert(text, len, "big5", "UTF-8", &rn, &wn, &err);
-
-    if (err) {
-      dbg("utf8 -> gb  convert error %d %d\n", rn, wn);
-      return;
-    }
-
-    err = NULL;
-    char *gbtext = g_convert(big5, wn, "GB2312", "big5", &rn, &wn, &err);
-    g_free(big5);
-
-    if (err) {
-      dbg("big5 -> gb  convert error %d %d\n", rn, wn);
-      return;
-    }
-
-    err = NULL;
-    utf8_gbtext = g_convert(gbtext, wn, "UTF-8", "GB2312", &rn, &wn, &err);
-    g_free(gbtext);
-
-    if (err) {
-      dbg("gb -> utf8  convert error %d %d\n", rn, wn);
-      return;
-    }
-
+    len = trad2sim(text, len, &utf8_gbtext);
     text = utf8_gbtext;
-    len = wn;
   }
 
-  int requiredN = len + 1 + output_bufferN;
+direct:
+  filter = getenv("GCIN_OUTPUT_FILTER");
+  char filter_text[512];
 
-  output_buffer = realloc(output_buffer, requiredN);
-  output_buffer[output_bufferN] = 0;
+  if (filter) {
+    int pfdr[2], pfdw[2];
 
-  strcat(output_buffer, text);
-  output_bufferN += len;
+    if (pipe(pfdr) == -1) {
+      dbg("cannot pipe r\n");
+      goto next;
+    }
 
-  g_free(utf8_gbtext);
+    if (pipe(pfdw) == -1) {
+      dbg("cannot pipe w\n");
+      goto next;
+    }
+
+    int pid = fork();
+
+    if (pid < 0) {
+      dbg("cannot fork filter\n");
+      goto next;
+    }
+
+    if (pid) {
+      close(pfdw[0]);
+      close(pfdr[1]);
+      write(pfdw[1], text, len);
+      close(pfdw[1]);
+      int rn = read(pfdr[0], filter_text, sizeof(filter_text) - 1);
+      filter_text[rn] = 0;
+//      puts(filter_text);
+      close(pfdr[0]);
+      text = filter_text;
+      len = rn;
+    } else {
+      close(pfdr[0]);
+      close(pfdw[1]);
+      dup2(pfdw[0], 0);
+      dup2(pfdr[1], 1);
+      if (execl(filter, filter, NULL) < 0) {
+        dbg("execl %s err", filter);
+        goto next;
+      }
+    }
+
+  }
+
+next:
+  if (len)
+    append_str(&output_buffer, &output_bufferN, text, len);
+
+  free(utf8_gbtext);
 }
 
 void send_output_buffer_bak()
 {
-  send_text(output_buffer_bak);
+  send_text(output_buffer_raw_bak);
 }
 
 void sendkey_b5(char *bchar)
@@ -120,6 +162,12 @@ void sendkey_b5(char *bchar)
   send_text(tt);
 }
 
+
+void send_ascii(char key)
+{
+  sendkey_b5(&key);
+}
+
 #if USE_XIM
 void export_text_xim()
 {
@@ -128,20 +176,12 @@ void export_text_xim()
   if (!output_bufferN)
     return;
 
-
   XTextProperty tp;
   char outbuf[512];
 
-#if !SOL
-  if (pxim_arr->b_send_utf8_str) {
-    Xutf8TextListToTextProperty(dpy, &text, 1, XCompoundTextStyle, &tp);
-  } else
-#endif
-  {
     utf8_big5(output_buffer, outbuf);
     text = outbuf;
     XmbTextListToTextProperty(dpy, &text, 1, XCompoundTextStyle, &tp);
-  }
 
 #if DEBUG
   dbg("sendkey_b5: %s\n", text);
@@ -197,19 +237,26 @@ void hide_in_win(ClientState *cs)
     case 3:
       hide_win_pho();
       break;
+#if USE_TSIN
     case 6:
 //      flush_tsin_buffer();
       hide_win0();
       break;
-    case 0:
+#endif
+    case 10:
       hide_win_int();
       break;
+    case 12:
+      hide_win_anthy();
+      break;
     default:
       hide_win_gtab();
   }
 
   reset_current_in_win_xy();
+#if 0
   hide_win_status();
+#endif
 }
 
 void show_win_pho();
@@ -217,6 +264,8 @@ void show_win0();
 void show_win_int();
 void show_win_gtab();
 
+gboolean force_show;
+
 void show_in_win(ClientState *cs)
 {
   if (!cs) {
@@ -230,17 +279,23 @@ void show_in_win(ClientState *cs)
     case 3:
       show_win_pho();
       break;
+#if USE_TSIN
     case 6:
       show_win0();
       break;
-    case 0:
+#endif
+    case 10:
       show_win_int();
       break;
+    case 12:
+      show_win_anthy();
+      break;
     default:
       show_win_gtab();
   }
-
+#if 0
   show_win_stautus();
+#endif
 }
 
 
@@ -276,12 +331,17 @@ void move_in_win(ClientState *cs, int x, int y)
     case 3:
       move_win_pho(x, y);
       break;
+#if USE_TSIN
     case 6:
       move_win0(x, y);
       break;
-    case 0:
+#endif
+    case 10:
       move_win_int(x, y);
       break;
+    case 12:
+      move_win_anthy(x, y);
+      break;
     default:
       if (!cs->in_method)
         return;
@@ -321,9 +381,11 @@ void move_IC_in_win(ClientState *cs)
 
    int inpx = cs->spot_location.x;
    int inpy = cs->spot_location.y;
-   XWindowAttributes att;
 
+   XWindowAttributes att;
    XGetWindowAttributes(dpy, inpwin, &att);
+   if (att.override_redirect)
+     return;
 
    if (inpx >= att.width)
      inpx = att.width - 1;
@@ -379,9 +441,11 @@ void win_pho_disp_half_full();
 void win_tsin_disp_half_full();
 void win_gtab_disp_half_full();
 extern char eng_full_str[], full_char_str[];
+void update_tray_icon();
 
 void disp_im_half_full()
 {
+#if 0
   if (current_CS->im_state == GCIN_STATE_ENG_FULL) {
      set_win_status_half_full(eng_full_str);
   }
@@ -390,14 +454,20 @@ void disp_im_half_full()
      set_win_status_half_full(current_CS->b_half_full_char?full_char_str:"");
   else
      set_win_status_half_full("");
+#endif
+#if TRAY_ENABLED
+  update_tray_icon();
+#endif
 
   switch (current_CS->in_method) {
     case 3:
       win_pho_disp_half_full();
       break;
+#if USE_TSIN
     case 6:
       win_tsin_disp_half_full();
       break;
+#endif
     default:
       win_gtab_disp_half_full();
       break;
@@ -407,23 +477,31 @@ void disp_im_half_full()
 gboolean flush_tsin_buffer();
 void reset_gtab_all();
 
+static u_int orig_caps_state;
+
+void init_state_chinese(ClientState *cs)
+{
+  cs->im_state = GCIN_STATE_CHINESE;
+  if (!cs->in_method)
+    init_in_method(default_input_method);
+}
+
 void toggle_im_enabled(u_int kev_state)
 {
 //    dbg("toggle_im_enabled\n");
     check_CS();
 
-    if (current_CS->in_method < 0 || current_CS->in_method >= gcin_switch_keysN)
-      return;
+    if (current_CS->in_method < 0 || current_CS->in_method > MAX_GTAB_NUM_KEY)
+      p_err("err found");
 
-    static u_int orig_caps_state;
 
     if (current_CS->im_state != GCIN_STATE_DISABLED) {
+#if 0
       if (current_CS->in_method== 6 && (kev_state & LockMask) != orig_caps_state &&
           tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_CapsLock) {
-        KeyCode caps = XKeysymToKeycode(dpy, XK_Caps_Lock);
-        XTestFakeKeyEvent(dpy, caps, True, CurrentTime);
-        XTestFakeKeyEvent(dpy, caps, False, CurrentTime);
+        send_fake_key_eve(XK_Caps_Lock);
       }
+#endif
 
       if (current_CS->im_state == GCIN_STATE_ENG_FULL) {
         current_CS->im_state = GCIN_STATE_CHINESE;
@@ -431,22 +509,29 @@ void toggle_im_enabled(u_int kev_state)
         return;
       }
 
-      if (current_CS->in_method == 6)
+      if (current_CS->in_method == 6) {
+#if USE_TSIN
         flush_tsin_buffer();
+#endif
+      }
       else {
+        output_gbuf();
         reset_gtab_all();
       }
 
       hide_in_win(current_CS);
+#if 0
       hide_win_status();
+#endif
       current_CS->im_state = GCIN_STATE_DISABLED;
-    } else {
-      current_CS->im_state = GCIN_STATE_CHINESE;
-      orig_caps_state = kev_state & LockMask;
 
-      if (!current_CS->in_method) {
-        init_in_method(default_input_method);
-      }
+#if TRAY_ENABLED
+      update_tray_icon();
+#endif
+    } else {
+      init_state_chinese(current_CS);
+
+      orig_caps_state = kev_state & LockMask;
 
       reset_current_in_win_xy();
 #if 1
@@ -457,8 +542,9 @@ void toggle_im_enabled(u_int kev_state)
       show_in_win(current_CS);
 #endif
 
-      if (gcin_pop_up_win)
-        show_win_stautus();
+#if TRAY_ENABLED
+      update_tray_icon();
+#endif
     }
 }
 
@@ -473,10 +559,12 @@ void update_active_in_win_geom()
     case 3:
       get_win_pho_geom();
       break;
+#if USE_TSIN
     case 6:
       get_win0_geom();
       break;
-    case 0:
+#endif
+    case 10:
       break;
     default:
       get_win_gtab_geom();
@@ -484,6 +572,30 @@ void update_active_in_win_geom()
   }
 }
 
+extern GtkWidget *gwin_pho, *gwin0, *gwin_gtab;
+
+gboolean win_is_visible()
+{
+  if (!current_CS)
+    return FALSE;
+  switch (current_CS->in_method) {
+    case 3:
+      return GTK_WIDGET_VISIBLE(gwin_pho);
+#if USE_TSIN
+    case 6:
+      return GTK_WIDGET_VISIBLE(gwin0);
+      break;
+#endif
+    case 10:
+      break;
+    default:
+      if (!gwin_gtab)
+        return FALSE;
+      return GTK_WIDGET_VISIBLE(gwin_gtab);
+  }
+}
+
+
 void disp_gtab_half_full(gboolean hf);
 void tsin_toggle_half_full();
 extern char eng_full_str[];
@@ -517,11 +629,12 @@ void toggle_half_full_char(u_int kev_state)
 //  dbg("half full toggle\n");
 }
 
-void init_gtab(int inmdno, int usenow);
-void init_inter_code(int usenow);
-void init_tab_pp(int usenow);
+void init_gtab(int inmdno);
+void init_inter_code();
+void init_tab_pp();
 void init_tab_pho();
 
+
 void check_CS()
 {
   if (!current_CS) {
@@ -533,10 +646,13 @@ void check_CS()
     temp_CS = *current_CS;
 }
 
+extern int b_show_win_kbm;
 
+void load_tray_icon();
+extern gboolean win_sym_enabled;
 gboolean init_in_method(int in_no)
 {
-  gboolean status = TRUE;
+  gboolean init_im = !(cur_inmd && (cur_inmd->flag & FLAG_GTAB_SYM_KBM));
 
   if (in_no < 0 || in_no > MAX_GTAB_NUM_KEY)
     return FALSE;
@@ -545,38 +661,72 @@ gboolean init_in_method(int in_no)
 
 
   if (current_CS->in_method != in_no) {
-    if (current_CS->in_method == 6)
+    if (!(inmd[in_no].flag & FLAG_GTAB_SYM_KBM)) {
+      if (current_CS->in_method == 6) {
+#if USE_TSIN
       flush_tsin_buffer();
+#endif
+      } else
+        output_gbuf();
 
-    hide_in_win(current_CS);
+      hide_in_win(current_CS);
+    }
+
+    if (cur_inmd && (cur_inmd->flag & FLAG_GTAB_SYM_KBM))
+      hide_win_kbm();
   }
 
   reset_current_in_win_xy();
 
 //  dbg("switch init_in_method %x %d\n", current_CS, in_no);
-  current_CS->in_method = in_no;
-
 
   switch (in_no) {
     case 3:
+      current_CS->in_method = in_no;
       init_tab_pho();
       break;
+#if USE_TSIN
     case 6:
-      init_tab_pp(True);
+      current_CS->in_method = in_no;
+      init_tab_pp(init_im);
       break;
-    case 0:
-      init_inter_code(True);
+#endif
+    case 10:
+      current_CS->in_method = in_no;
+      init_inter_code();
+      break;
+    case 12:
+      if (init_win_anthy()) {
+        current_CS->in_method = in_no;
+        show_win_anthy();
+      } else {
+        return FALSE;
+      }
       break;
     default:
-      show_win_gtab();
-//      init_gtab(in_no, True);
+      init_gtab(in_no);
+      if (!inmd[in_no].DefChars)
+        return FALSE;
+      current_CS->in_method = in_no;
+      if (!(inmd[in_no].flag & FLAG_GTAB_SYM_KBM))
+        show_win_gtab();
+      else {
+        win_kbm_inited = 1;
+        show_win_kbm();
+      }
       break;
   }
 
-  set_win_status_inmd(inmd[in_no].cname);
+#if TRAY_ENABLED
+  load_tray_icon();
+#endif
 
   update_in_win_pos();
-  return status;
+
+  if (win_kbm_inited)
+    update_win_kbm();
+
+  return TRUE;
 }
 
 
@@ -585,14 +735,16 @@ static void cycle_next_in_method()
 
   int i;
 
-  for(i=1; i <= gcin_switch_keysN; i++) {
-    int v = (current_CS->in_method + i) % gcin_switch_keysN;
+  for(i=0; i < MAX_GTAB_NUM_KEY; i++) {
+    int v = ((current_CS->in_method + i) % MAX_GTAB_NUM_KEY) + 1;
     if (!(gcin_flags_im_enabled & (1<<v)))
       continue;
     if (!inmd[v].cname || !inmd[v].cname[0])
       continue;
 
-    init_in_method(v);
+    if (!init_in_method(v))
+      continue;
+
     return;
   }
 }
@@ -613,6 +765,9 @@ gboolean full_char_proc(KeySym keysym)
 }
 
 
+
+
+
 int feedkey_pho(KeySym xkey, int kbstate);
 int feedkey_pp(KeySym xkey, int state);
 int feedkey_gtab(KeySym key, int kbstate);
@@ -623,24 +778,24 @@ static KeySym last_keysym;
 
 gboolean timeout_raise_window()
 {
+//  dbg("timeout_raise_window\n");
   timeout_handle = 0;
   show_in_win(current_CS);
   return FALSE;
 }
 
-
 // return TRUE if the key press is processed
 gboolean ProcessKeyPress(KeySym keysym, u_int kev_state)
 {
 #if 0
-  dbg_time("key press %x %x", keysym, kev_state);
+  dbg_time("key press %x %x\n", keysym, kev_state);
 #endif
   check_CS();
 
   if (current_CS->client_win)
     focus_win = current_CS->client_win;
 
-  if (strlen(callback_str_buffer)) {
+  if (callback_str_buffer && strlen(callback_str_buffer)) {
     send_text(callback_str_buffer);
     callback_str_buffer[0]=0;
     return TRUE;
@@ -652,7 +807,7 @@ gboolean ProcessKeyPress(KeySym keysym, u_int kev_state)
     dbg("%x\n", Mod4Mask);
 #endif
     if (
-      ((kev_state & ControlMask) && gcin_im_toggle_keys==Control_Space) ||
+      ((kev_state & (ControlMask|Mod1Mask|ShiftMask))==ControlMask && gcin_im_toggle_keys==Control_Space) ||
       ((kev_state & Mod1Mask) && gcin_im_toggle_keys==Alt_Space) ||
       ((kev_state & ShiftMask) && gcin_im_toggle_keys==Shift_Space) ||
       ((kev_state & Mod4Mask) && gcin_im_toggle_keys==Windows_Space)
@@ -677,7 +832,10 @@ gboolean ProcessKeyPress(KeySym keysym, u_int kev_state)
 
 
   if ((kev_state & (Mod1Mask|ShiftMask)) == (Mod1Mask|ShiftMask)) {
-    return feed_phrase(keysym, kev_state);
+    if (current_CS->im_state != GCIN_STATE_DISABLED || gcin_eng_phrase_enabled)
+      return feed_phrase(keysym, kev_state);
+    else
+      return 0;
   }
 
 //  dbg("state %x\n", kev_state);
@@ -693,13 +851,27 @@ gboolean ProcessKeyPress(KeySym keysym, u_int kev_state)
     }
 
     if (keysym == ',') {
-      create_win_sym();
 
+      if (current_CS->im_state == GCIN_STATE_CHINESE) {
+        if (!win_is_visible())
+          win_sym_enabled=1;
+        else
+          win_sym_enabled^=1;
+      } else
+        win_sym_enabled==0;
+
+      create_win_sym();
+      if (win_sym_enabled) {
+        force_show = TRUE;
+        if (current_CS->im_state == GCIN_STATE_CHINESE)
+          show_in_win(current_CS);
+        force_show = FALSE;
+      }
       return TRUE;
     }
 
     int kidx = gcin_switch_keys_lookup(keysym);
-    if (kidx < 0)
+    if (kidx <= 0)
       return FALSE;
 
     current_CS->im_state = GCIN_STATE_CHINESE;
@@ -714,6 +886,7 @@ gboolean ProcessKeyPress(KeySym keysym, u_int kev_state)
   }
 
 
+  if (!current_CS->b_gcin_protocol) {
   if (((keysym == XK_Control_L || keysym == XK_Control_R)
                    && (kev_state & ShiftMask)) ||
       ((keysym == XK_Shift_L || keysym == XK_Shift_R)
@@ -721,18 +894,30 @@ gboolean ProcessKeyPress(KeySym keysym, u_int kev_state)
      cycle_next_in_method();
      return TRUE;
   }
+  }
 
-  if (current_CS->b_raise_window && !timeout_handle) {
+  if (current_CS->b_raise_window && keysym>=' ' && keysym < 127) {
+    if (timeout_handle)
+      g_source_remove(timeout_handle);
     timeout_handle = g_timeout_add(200, timeout_raise_window, NULL);
   }
 
+  if (kev_state & ControlMask) {
+    if (feed_phrase(keysym, kev_state))
+      return TRUE;
+  }
+
   switch(current_CS->in_method) {
     case 3:
       return feedkey_pho(keysym, kev_state);
+#if USE_TSIN
     case 6:
       return feedkey_pp(keysym, kev_state);
-    case 0:
+#endif
+    case 10:
       return feedkey_intcode(keysym);
+    case 12:
+      return feedkey_anthy(keysym, kev_state);
     default:
       return feedkey_gtab(keysym, kev_state);
   }
@@ -747,14 +932,33 @@ int feedkey_pp_release(KeySym xkey, int kbstate);
 gboolean ProcessKeyRelease(KeySym keysym, u_int kev_state)
 {
   check_CS();
+#if 0
+  dbg_time("key release %x %x\n", keysym, kev_state);
+#endif
 
   if (current_CS->im_state == GCIN_STATE_DISABLED)
     return FALSE;
 
+#if 1
+  if (current_CS->b_gcin_protocol && (last_keysym == XK_Shift_L ||
+  last_keysym == XK_Shift_R || last_keysym == XK_Control_L || last_keysym == XK_Control_R)) {
+    if (((keysym == XK_Control_L || keysym == XK_Control_R)
+          && (kev_state & ShiftMask)) ||
+        ((keysym == XK_Shift_L || keysym == XK_Shift_R)
+          && (kev_state & ControlMask))) {
+       cycle_next_in_method();
+       return TRUE;
+    }
+  }
+#endif
+
+
+#if USE_TSIN
   switch(current_CS->in_method) {
     case 6:
       return feedkey_pp_release(keysym, kev_state);
   }
+#endif
 
   return FALSE;
 }
@@ -789,13 +993,28 @@ int xim_ForwardEventHandler(IMForwardEventStruct *call_data)
 }
 #endif
 
-IC *FindIC(CARD16 icid);
-void load_IC(IC *rec);
+int skip_window(Window win)
+{
+  XWindowAttributes att;
+  XGetWindowAttributes(dpy, win, &att);
+#if 0
+  dbg("hhh %d %d class:%d all:%x %x\n", att.width, att.height, att.class,
+    att.all_event_masks, att.your_event_mask);
+#endif
+  if (att.override_redirect)
+    return 1;
+
+  return 0;
+}
+
 
 int gcin_FocusIn(ClientState *cs)
 {
   Window win = cs->client_win;
 
+  if (skip_window(win))
+    return;
+
   reset_current_in_win_xy();
 
   if (cs) {
@@ -809,6 +1028,8 @@ int gcin_FocusIn(ClientState *cs)
 
   current_CS = cs;
 
+//  dbg("current_CS %x %d %d\n", cs, cs->im_state, current_CS->im_state);
+
   if (win == focus_win) {
     if (cs->im_state != GCIN_STATE_DISABLED) {
 #if 0
@@ -821,12 +1042,16 @@ int gcin_FocusIn(ClientState *cs)
       show_in_win(cs);
       move_IC_in_win(cs);
 #endif
+#if 0
       set_win_status_inmd(inmd[cs->in_method].cname);
+#endif
     } else
       hide_in_win(cs);
   }
 
-
+#if TRAY_ENABLED
+  load_tray_icon();
+#endif
 #if 0
   dbg_time("gcin_FocusIn %x %x\n",cs, current_CS);
 #endif
@@ -835,6 +1060,8 @@ int gcin_FocusIn(ClientState *cs)
 
 
 #if USE_XIM
+IC *FindIC(CARD16 icid);
+void load_IC(IC *rec);
 int xim_gcin_FocusIn(IMChangeFocusStruct *call_data)
 {
     IC *ic = FindIC(call_data->icid);
@@ -853,18 +1080,40 @@ int xim_gcin_FocusIn(IMChangeFocusStruct *call_data)
 }
 #endif
 
+extern Window xwin_pho, xwin0, xwin_gtab;
+
+gint64 current_time();
+
+static gint64 last_focus_out_time;
+
 int gcin_FocusOut(ClientState *cs)
 {
-    if (cs == current_CS) {
-      hide_in_win(cs);
-    }
+  gint64 t = current_time();
 
-    reset_current_in_win_xy();
+  if (skip_window(cs->client_win))
+    return;
 
-    if (cs == current_CS)
-      temp_CS = *current_CS;
+  if (t - last_focus_out_time < 100000) {
+    last_focus_out_time = t;
+    return;
+  }
 
-    return True;
+  last_focus_out_time = t;
+
+  if (cs == current_CS) {
+    hide_in_win(cs);
+  }
+
+  reset_current_in_win_xy();
+
+  if (cs == current_CS)
+    temp_CS = *current_CS;
+
+#if 0
+  dbg("focus out\n");
+#endif
+
+  return True;
 }
 
 #if USE_XIM
