@@ -1,6 +1,6 @@
 
 /*
-	Copyright (C) 1995	Edward Der-Hua Liu, Hsin-Chu, Taiwan
+	Copyright (C) 1995-2008	Edward Der-Hua Liu, Hsin-Chu, Taiwan
 */
 
 #include <stdio.h>
@@ -13,10 +13,11 @@
 #include <string.h>
 #include "gcin.h"
 #include "gtab.h"
-
+#include "gcin-endian.h"
 
 FILE *fr, *fw;
 int lineno;
+char tt[512];
 
 
 char *skip_spc(char *s)
@@ -47,9 +48,10 @@ void del_nl_spc(char *s)
 }
 
 
-void get_line(char *tt)
+void get_line()
 {
   while (!feof(fr)) {
+    bzero(tt, sizeof(tt));
     fgets((char *)tt, 512, fr);
     lineno++;
 
@@ -64,18 +66,18 @@ void get_line(char *tt)
   }
 }
 
-void cmd_arg(char *s, char **cmd, char **arg)
+void cmd_arg(u_char **cmd, u_char **arg)
 {
   char *t;
 
-  get_line(s);
+  get_line();
+  char *s=tt;
 
   if (!*s) {
     *cmd=*arg=s;
     return;
   }
 
-
   s=skip_spc(s);
   t=to_spc(s);
   *cmd=s;
@@ -117,38 +119,19 @@ typedef struct {
 
 #define MAX_K (500000)
 
-ITEM2 itar[MAX_K], itmp[MAX_K];
-ITEM2_64 itar64[MAX_K], itmp64[MAX_K];
+ITEM2 itar[MAX_K];
+ITEM2_64 itar64[MAX_K];
 
 ITEM itout[MAX_K];
 ITEM64 itout64[MAX_K];
 
-int qcmp2(const void *aa, const void *bb)
-{
-  ITEM2 *a = (ITEM2 *)aa, *b = (ITEM2 *) bb;
-
-  u_int d;
-  if ((d = a->key - b->key)) return d;
-
-  return memcmp(a->ch ,b->ch, CH_SZ);
-}
-
-int qcmp2_64(const void *aa, const void *bb)
-{
-  ITEM2_64 *a = (ITEM2_64 *)aa, *b = (ITEM2_64 *) bb;
-
-  if (a->key > b->key) return 1;
-  if (a->key < b->key) return -1;
-
-  return memcmp(a->ch ,b->ch, CH_SZ);
-}
 
 int qcmp(const void *aa, const void *bb)
 {
   ITEM2 *a = (ITEM2 *)aa, *b = (ITEM2 *) bb;
 
-  u_int d;
-  if ((d = a->key - b->key)) return d;
+  if (a->key > b->key) return 1;
+  if (a->key < b->key) return -1;
 
   return a->oseq - b->oseq;
 }
@@ -175,13 +158,12 @@ int main(int argc, char **argv)
   char fname[64];
   char fname_cin[64];
   char fname_tab[64];
-  char tt[512];
   u_char *cmd, *arg;
   struct TableHead th;
   int KeyNum;
   char kname[128][CH_SZ];
-  char keymap[64];
-  int chno,cpcount;
+  char keymap[128];
+  int chno;
   gtab_idx1_t idx1[256];
   char def1[256];
   int quick_def;
@@ -203,10 +185,8 @@ int main(int argc, char **argv)
     exit(0);
   }
 
-
   char *p;
-
-  if((p=strchr(fname, '.')))
+  if(p=strstr(fname, ".cin"))
     *p = 0;
 
   strcpy(fname_cin,fname);
@@ -226,10 +206,10 @@ int main(int argc, char **argv)
   bzero(itar64,sizeof(itar64));
   bzero(itout64,sizeof(itout64));
 
-  cmd_arg(tt, &cmd, &arg);
+  cmd_arg(&cmd, &arg);
   if (sequ(cmd, "%gen_inp") || sequ(cmd, "%encoding")) {
     dbg("skip gen_inp\n");
-    cmd_arg(tt, &cmd, &arg);
+    cmd_arg(&cmd, &arg);
   }
 
   if (!sequ(cmd,"%ename") || !(*arg) )
@@ -237,39 +217,49 @@ int main(int argc, char **argv)
   arg[15]=0;
 //  strcpy(th.ename,arg);
 
-  cmd_arg(tt, &cmd, &arg);
+  cmd_arg(&cmd, &arg);
   if (!(sequ(cmd,"%prompt") || sequ(cmd,"%cname")) || !(*arg) )
     p_err("%d:  %%prompt prompt_name  expected", lineno);
   strncpy(th.cname, arg, MAX_CNAME);
   printf("cname %s\n", th.cname);
 
-  cmd_arg(tt,&cmd, &arg);
+  cmd_arg(&cmd, &arg);
   if (!sequ(cmd,"%selkey") || !(*arg) )
     p_err("%d:  %%selkey select_key_list expected", lineno);
   strcpy(th.selkey,arg);
 
-  cmd_arg(tt,&cmd, &arg);
+  cmd_arg(&cmd, &arg);
   if (!sequ(cmd,"%dupsel") || !(*arg) ) {
     th.M_DUP_SEL = strlen(th.selkey);
   }
   else {
     th.M_DUP_SEL=atoi(arg);
-    cmd_arg(tt,&cmd, &arg);
+    cmd_arg(&cmd, &arg);
   }
 
   if (sequ(cmd,"%endkey")) {
     strcpy(th.endkey, arg);
-    cmd_arg(tt,&cmd, &arg);
+    cmd_arg(&cmd, &arg);
   }
 
   if (sequ(cmd,"%space_style")) {
     th.space_style = atoi(arg);
-    cmd_arg(tt,&cmd, &arg);
+    cmd_arg(&cmd, &arg);
   }
 
   if (sequ(cmd,"%keep_key_case")) {
     th.flag |= FLAG_KEEP_KEY_CASE;
-    cmd_arg(tt,&cmd, &arg);
+    cmd_arg(&cmd, &arg);
+  }
+
+  if (sequ(cmd,"%symbol_kbm")) {
+    th.flag |= FLAG_GTAB_SYM_KBM;
+    cmd_arg(&cmd, &arg);
+  }
+
+  if (sequ(cmd,"%phase_auto_skip_endkey")) {
+    th.flag |= FLAG_PHRASE_AUTO_SKIP_ENDKEY;
+    cmd_arg(&cmd, &arg);
   }
 
   if (!sequ(cmd,"%keyname") || !sequ(arg,"begin")) {
@@ -279,7 +269,7 @@ int main(int argc, char **argv)
   for(KeyNum=0;;) {
     char k;
 
-    cmd_arg(tt,&cmd, &arg);
+    cmd_arg(&cmd, &arg);
     if (sequ(cmd,"%keyname")) break;
     if (BITON(th.flag, FLAG_KEEP_KEY_CASE))
       k=cmd[0];
@@ -298,7 +288,7 @@ int main(int argc, char **argv)
   KeyNum++;
   th.KeyS=KeyNum;    /* include space */
 
-  cmd_arg(tt,&cmd, &arg);
+  cmd_arg(&cmd, &arg);
 
   if (sequ(cmd,"%quick") && sequ(arg,"begin")) {
     dbg(".. quick keys defined\n");
@@ -306,7 +296,7 @@ int main(int argc, char **argv)
       char k;
       int len;
 
-      cmd_arg(tt,&cmd, &arg);
+      cmd_arg(&cmd, &arg);
       if (sequ(cmd,"%quick")) break;
       k=kno[mtolower(cmd[0])]-1;
 
@@ -342,15 +332,19 @@ int main(int argc, char **argv)
   int olineno = lineno;
   gboolean key64 = FALSE;
 
-
   while (!feof(fr)) {
     int len;
 
-    cmd_arg(tt,&cmd,&arg);
+    cmd_arg(&cmd,&arg);
     if (!cmd[0] || !arg[0])
       continue;
-    if (cmd[0]=='%')
-      continue;
+
+    if (!strcmp(cmd, "%chardef")) {
+      if (!strcmp(arg, "end"))
+        break;
+      else
+        continue;
+    }
 
     len=strlen(cmd);
 
@@ -366,11 +360,21 @@ int main(int argc, char **argv)
 
   INMD inmd, *cur_inmd = &inmd;
 
-  cpcount=0;
   cur_inmd->key64 = key64;
   cur_inmd->tbl64 = itout64;
   cur_inmd->tbl = itout;
 
+  if (KeyNum < 64)
+    cur_inmd->keybits = 6;
+  else
+    cur_inmd->keybits = 7;
+
+  dbg("KeyNum:%d keybits:%d\n", KeyNum, cur_inmd->keybits);
+
+  th.keybits = cur_inmd->keybits;
+  cur_inmd->last_k_bitn = (((cur_inmd->key64 ? 64:32) / cur_inmd->keybits) - 1) * cur_inmd->keybits;
+
+
   puts("char def");
   chno=0;
   while (!feof(fr)) {
@@ -378,41 +382,52 @@ int main(int argc, char **argv)
     u_int64_t kk;
     int k;
 
-    cmd_arg(tt, &cmd, &arg);
+    cmd_arg(&cmd, &arg);
     if (!cmd[0] || !arg[0])
       continue;
-    if (cmd[0]=='%')
-      continue;
+
+    if (!strcmp(cmd, "%chardef")) {
+      if (!strcmp(arg, "end"))
+        break;
+      else
+        continue;
+    }
 
     len=strlen(cmd);
 
-    if (len > th.MaxPress)
+    if (len > th.MaxPress) {
       th.MaxPress=len;
+    }
 
     if (len > 10)
       p_err("%d:  only <= 10 keys is allowed '%s'", lineno, cmd);
 
     kk=0;
     for(i=0;i<len;i++) {
-      int key =  BITON(th.flag, FLAG_KEEP_KEY_CASE) ?
+      int key = BITON(th.flag, FLAG_KEEP_KEY_CASE) ?
         cmd[i] : mtolower(cmd[i]);
 
       k=kno[key];
-      kk|=(u_int64_t)k << ( LAST_K_bitN - i*6);
+
+      if (!k)
+        p_err("%d: key undefined in keyname '%c'\n", lineno, cmd[i]);
+
+      kk|=(u_int64_t)k << ( LAST_K_bitN - i*th.keybits);
     }
 
-    if (key64)
+    if (key64) {
       memcpy(&itar64[chno].key, &kk, 8);
+      itar64[chno].oseq=chno;
+    }
     else {
       u_int key32 = kk;
 
       memcpy(&itar[chno].key, &key32, 4);
+      itar[chno].oseq=chno;
     }
 
-
     if ((len=strlen(arg)) <= CH_SZ && (arg[0] & 0x80)) {
       char out[CH_SZ+1];
-      int u8len = utf8_sz(arg);
 
       bzero(out, sizeof(out));
       memcpy(out, arg, len);
@@ -435,7 +450,7 @@ int main(int argc, char **argv)
       }
 
       if (len > MAX_CIN_PHR)
-        p_err("phrase too long: %s\n", arg);
+        p_err("phrase too long: %s  max:%d bytes\n", arg, MAX_CIN_PHR);
 
       phridx = trealloc(phridx, int, phr_cou+1);
       phridx[phr_cou++]=prbf_cou;
@@ -445,59 +460,39 @@ int main(int argc, char **argv)
       prbf_cou+=len;
     }
 
-    itar[chno].oseq=chno;
     chno++;
   }
   fclose(fr);
 
-  if (key64)
-    qsort(itar, chno,sizeof(ITEM2), qcmp2_64);
-  else
-    qsort(itar, chno,sizeof(ITEM2), qcmp2);
+#define _sort qsort
 
+  dbg("MaxPress: %d\n", th.MaxPress);
 
-
-  if (key64) {
-    for(i=0;i<chno;i++) {
-      if (!i || memcmp(&itar64[i], &itar64[i-1], sizeof(ITEM64))) {
-        memcpy(&itmp64[cpcount++], &itar64[i], sizeof(ITEM2_64));
-      }
-    }
-  } else {
-    for(i=0;i<chno;i++) {
-      if (!i || memcmp(&itar[i], &itar[i-1], sizeof(ITEM))) {
-        memcpy(&itmp[cpcount++],&itar[i],sizeof(ITEM2));
-      }
-    }
-  }
-
-  chno=cpcount;
   th.DefC=chno;
   cur_inmd->DefChars = chno;
 
   if (key64)
-    qsort(itmp64,chno,sizeof(ITEM2_64),qcmp_64);
+    _sort(itar64,chno,sizeof(ITEM2_64),qcmp_64);
   else
-    qsort(itmp,chno,sizeof(ITEM2),qcmp);
-
+    _sort(itar,chno,sizeof(ITEM2),qcmp);
 
   if (key64) {
-    for(i=0;i<chno;i++) {
-      memcpy(&itout64[i],&itmp64[i],sizeof(ITEM64));
-    }
+    for(i=0;i<chno;i++)
+      memcpy(&itout64[i],&itar64[i],sizeof(ITEM64));
   } else {
-    for(i=0;i<chno;i++) {
-      memcpy(&itout[i],&itmp[i],sizeof(ITEM));
-    }
+    for(i=0;i<chno;i++)
+      memcpy(&itout[i],&itar[i],sizeof(ITEM));
   }
 
 
   bzero(def1,sizeof(def1));
   bzero(idx1,sizeof(idx1));
 
+
+  u_int64_t keymask = KEY_MASK;
   for(i=0; i<chno; i++) {
     u_int64_t key = CONVT2(cur_inmd, i);
-    int kk = (key>>LAST_K_bitN) & 0x3f;
+    int kk = (key>>LAST_K_bitN) & keymask;
 
     if (!def1[kk]) {
       idx1[kk]=(gtab_idx1_t)i;
@@ -515,26 +510,57 @@ int main(int argc, char **argv)
 
   printf("Defined Characters:%d\n", chno);
 
+#if NEED_SWAP
+  swap_byte_4(&th.version);
+  swap_byte_4(&th.flag);
+  swap_byte_4(&th.space_style);
+  swap_byte_4(&th.KeyS);
+  swap_byte_4(&th.MaxPress);
+  swap_byte_4(&th.M_DUP_SEL);
+  swap_byte_4(&th.DefC);
+  for(i=0; i <= KeyNum; i++)
+    swap_byte_4(&idx1[i]);
+#endif
   fwrite(&th,1,sizeof(th),fw);
   fwrite(keymap, 1, KeyNum, fw);
   fwrite(kname, CH_SZ, KeyNum, fw);
+
   fwrite(idx1, sizeof(gtab_idx1_t), KeyNum+1, fw);
 
+
   if (key64) {
+#if NEED_SWAP
+    for(i=0; i < chno; i++) {
+      swap_byte_8(&itout64[i].key);
+    }
+#endif
     fwrite(itout64, sizeof(ITEM64), chno, fw);
 #if 0
     for(i=0; i < 100; i++)
       dbg("%d] %c%c%c\n", i, itout64[i].ch[0], itout64[i].ch[1], itout64[i].ch[2]);
 #endif
   }
-  else
+  else {
+#if NEED_SWAP
+    for(i=0; i < chno; i++) {
+      swap_byte_4(&itout[i].key);
+    }
+#endif
     fwrite(itout, sizeof(ITEM), chno, fw);
+  }
 
   if (phr_cou) {
     phridx[phr_cou++]=prbf_cou;
     printf("phrase count:%d\n", phr_cou);
+
+    int ophr_cou = phr_cou;
+#if NEED_SWAP
+    for(i=0; i < phr_cou; i++)
+      swap_byte_4(&phridx[i]);
+    swap_byte_4(&phr_cou);
+#endif
     fwrite(&phr_cou, sizeof(int), 1, fw);
-    fwrite(phridx, sizeof(int), phr_cou, fw);
+    fwrite(phridx, sizeof(int), ophr_cou, fw);
     fwrite(phrbuf,1,prbf_cou,fw);
   }
 
