@@ -28,6 +28,10 @@
 #include "gtkimcontextgcin.h"
 // #include "gcin.h"  // for debug only
 #include "gcin-im-client.h"
+#include <X11/keysym.h>
+
+// #define NEW_GTK_IM 1
+#define DBG 0
 
 typedef struct _GtkGCINInfo GtkGCINInfo;
 
@@ -39,13 +43,16 @@ struct _GtkIMContextGCIN
   GtkWidget *client_widget;
   GCIN_client_handle *gcin_ch;
   int timeout_handle;
-  gboolean is_mozilla;
+  gboolean is_mozilla, dirty_fix_off;
 };
 
 
 #if NEW_GTK_IM
 static void cancel_timeout(GtkIMContextGCIN *context)
 {
+#if DBG
+  printf("cancel_timeout %d\n", context->timeout_handle);
+#endif
   if (!context->timeout_handle)
     return;
   g_source_remove(context->timeout_handle);
@@ -77,7 +84,7 @@ static void     gtk_im_context_gcin_get_preedit_string (GtkIMContext          *c
 GType gtk_type_im_context_gcin = 0;
 
 void
-gtk_im_context_xim_register_type (GTypeModule *type_module)
+gtk_im_context_gcin_register_type (GTypeModule *type_module)
 {
   static const GTypeInfo im_context_gcin_info =
   {
@@ -102,12 +109,19 @@ gtk_im_context_xim_register_type (GTypeModule *type_module)
 static void
 reinitialize_all_ics (GtkGCINInfo *info)
 {
+#if DBG
+  puts("reinitialize_all_ics");
+#endif
 }
 
+#if 0
 static void gcin_display_closed (GdkDisplay *display,
 			 gboolean    is_error,
                          GtkIMContextGCIN *context_xim)
 {
+#if DBG
+  puts("gcin_display_closed");
+#endif
 #if NEW_GTK_IM
   cancel_timeout(context_xim);
 #endif
@@ -117,6 +131,7 @@ static void gcin_display_closed (GdkDisplay *display,
   gcin_im_client_close(context_xim->gcin_ch);
   context_xim->gcin_ch = NULL;
 }
+#endif
 
 static void
 preedit_style_change (GtkGCINInfo *info)
@@ -136,13 +151,18 @@ get_im (GtkIMContextGCIN *context_xim)
   if (!context_xim->gcin_ch) {
     if (!(context_xim->gcin_ch = gcin_im_client_open(GDK_DISPLAY())))
       perror("cannot open gcin_ch");
-
+#if 0
+    // coredump
     g_signal_connect (display, "closed",
                       G_CALLBACK (gcin_display_closed), context_xim);
-
-    if (context_xim->is_mozilla)
+#endif
+    if (context_xim->is_mozilla) {
+      int rflag;
       gcin_im_client_set_flags(context_xim->gcin_ch,
-        FLAG_GCIN_client_handle_raise_window);
+        FLAG_GCIN_client_handle_raise_window, &rflag);
+
+      context_xim->dirty_fix_off = (rflag & FLAG_GCIN_srv_ret_status_use_pop_up) > 0;
+    }
   }
 }
 
@@ -169,11 +189,15 @@ static void
 gtk_im_context_gcin_init (GtkIMContextGCIN *im_context_gcin)
 {
   im_context_gcin->timeout_handle = 0;
-
+#if DBG
+  printf("gtk_im_context_gcin_init %x\n", im_context_gcin);
+#endif
 #if NEW_GTK_IM
   int pid = getpid();
 // probably only works for linux
-  static char *moz[]={"mozilla", "firefox", "thunderbird", "nvu"};
+  static char *moz[]={"mozilla", "firefox", "thunderbird", "nvu", "sunbird",
+	"seamonkey", "gnuzilla", "iceweasel", "icedove", "iceape", "swiftfox",
+	"iceowl", "kompozer", "swiftdove", "swiftweasel", "navigator", "xulrunner"};
   char tstr0[64];
   char exec[256];
   sprintf(tstr0, "/proc/%d/exe", pid);
@@ -195,16 +219,19 @@ gtk_im_context_gcin_init (GtkIMContextGCIN *im_context_gcin)
 static void
 gtk_im_context_gcin_finalize (GObject *obj)
 {
-//  printf("gtk_im_context_gcin_finalize\n");
+#if DBG
+  printf("gtk_im_context_gcin_finalize %x\n", obj);
+#endif
   GtkIMContextGCIN *context_xim = GTK_IM_CONTEXT_GCIN (obj);
 
+#if NEW_GTK_IM
+  cancel_timeout(context_xim);
+#endif
+
   if (context_xim->gcin_ch) {
     gcin_im_client_close(context_xim->gcin_ch);
     context_xim->gcin_ch = NULL;
   }
-#if NEW_GTK_IM
-  cancel_timeout(context_xim);
-#endif
 }
 
 /* Finds the GtkWidget that owns the window, or if none, the
@@ -228,13 +255,10 @@ widget_for_window (GdkWindow *window)
 }
 
 
-/* Updates the in_toplevel flag for @context_xim
- */
+#if 0
 static void update_in_toplevel (GtkIMContextGCIN *context_xim)
 {
 }
-
-
 static void
 on_client_widget_hierarchy_changed (GtkWidget       *widget,
 				    GtkWidget       *old_toplevel,
@@ -242,13 +266,18 @@ on_client_widget_hierarchy_changed (GtkWidget       *widget,
 {
   update_in_toplevel (context_xim);
 }
-
+#endif
 
 static void
 set_ic_client_window (GtkIMContextGCIN *context_xim,
                       GdkWindow       *client_window)
 {
-//  printf("set_ic_client_window\n");
+#if DBG
+  printf("set_ic_client_window %x %x\n", context_xim, client_window);
+#endif
+  if (!client_window)
+    return;
+
   context_xim->client_window = client_window;
 
   if (context_xim->client_window)
@@ -259,7 +288,8 @@ set_ic_client_window (GtkIMContextGCIN *context_xim,
       }
     }
 
-
+#if 0
+  // cause coredump
   GtkWidget *new_client_widget = widget_for_window (context_xim->client_window);
 
   if (new_client_widget != context_xim->client_widget)
@@ -280,6 +310,7 @@ set_ic_client_window (GtkIMContextGCIN *context_xim,
 
       update_in_toplevel (context_xim);
     }
+#endif
 }
 
 #if NEW_GTK_IM
@@ -288,7 +319,7 @@ static gboolean update_cursor_position(gpointer data)
   GtkIMContextGCIN *context = (GtkIMContextGCIN *)data;
 
   g_signal_emit_by_name(context, "preedit_changed");
-  context->timeout_handle = 0;
+  cancel_timeout(context);
   return FALSE;
 }
 #endif
@@ -308,11 +339,12 @@ static void
 gtk_im_context_gcin_set_client_window (GtkIMContext          *context,
 				      GdkWindow             *client_window)
 {
-//  printf("gtk_im_context_gcin_set_client_window\n");
-
+#if DBG
+  printf("gtk_im_context_gcin_set_client_window\n");
+#endif
   GtkIMContextGCIN *context_xim = GTK_IM_CONTEXT_GCIN (context);
 #if NEW_GTK_IM
-  if (context_xim->is_mozilla)
+  if (context_xim->is_mozilla && !context_xim->dirty_fix_off)
     add_cursor_timeout(context_xim);
 #endif
   set_ic_client_window (context_xim, client_window);
@@ -320,9 +352,11 @@ gtk_im_context_gcin_set_client_window (GtkIMContext          *context,
 
 ///
 GtkIMContext *
-gtk_im_context_xim_new (void)
+gtk_im_context_gcin_new (void)
 {
-//  printf("gtk_im_context_gcin_new\n");
+#if DBG
+  printf("gtk_im_context_gcin_new\n");
+#endif
   GtkIMContextGCIN *result;
 
   result = GTK_IM_CONTEXT_GCIN (g_object_new (GTK_TYPE_IM_CONTEXT_GCIN, NULL));
@@ -331,8 +365,6 @@ gtk_im_context_xim_new (void)
 }
 
 
-
-
 ///
 static gboolean
 gtk_im_context_gcin_filter_keypress (GtkIMContext *context,
@@ -341,10 +373,11 @@ gtk_im_context_gcin_filter_keypress (GtkIMContext *context,
   GtkIMContextGCIN *context_xim = GTK_IM_CONTEXT_GCIN (context);
 
   gchar static_buffer[256];
-  gchar *buffer = static_buffer;
+  unsigned char *buffer = static_buffer;
+//  char *buffer = static_buffer;
   gint buffer_size = sizeof(static_buffer) - 1;
   gint num_bytes = 0;
-  KeySym keysym;
+  KeySym keysym = 0;
   Status status;
   gboolean result = FALSE;
   GdkWindow *root_window = gdk_screen_get_root_window (gdk_drawable_get_screen (event->window));
@@ -368,15 +401,44 @@ gtk_im_context_gcin_filter_keypress (GtkIMContext *context,
   char *rstr = NULL;
   num_bytes = XLookupString (&xevent, buffer, buffer_size, &keysym, NULL);
 
+  // If it is latin key, XLookupString only works for UTF-8 env
+
+#if (!FREEBSD || MAC_OS)
+  char *lc_ctype = getenv("LC_CTYPE");
+  if (!lc_ctype || !strstr(lc_ctype, "UTF-8")) {
+    int uni = gdk_keyval_to_unicode(event->keyval);
+    if (uni) {
+      unsigned int rn;
+      GError *err = NULL;
+      char *utf8 = g_convert((char *)&uni, 4, "UTF-8", "UTF-32", &rn, &num_bytes, &err);
+
+      if (utf8) {
+        strcpy(buffer, utf8);
+        g_free(utf8);
+      }
+    }
+  }
+#endif
+
   if (xevent.type == KeyPress) {
     result = gcin_im_client_forward_key_press(context_xim->gcin_ch,
       keysym, xevent.state, &rstr);
 
 #if NEW_GTK_IM
-    // this one is for mozilla
-    if (context_xim->is_mozilla && (rstr || !result))
-      add_cursor_timeout(context_xim);
+    // this one is for mozilla, I know it is very dirty
+    if (context_xim->is_mozilla) {
+      if (context_xim->dirty_fix_off) {
+        if (result && !rstr)
+          g_signal_emit_by_name(context, "preedit_changed");
+      } else {
+        if (rstr || !result) {
+          add_cursor_timeout(context_xim);
+        }
+      }
+    }
 #endif
+
+//    printf("jj %x %x %d %x\n", rstr, result, num_bytes, (unsigned int)buffer[0]);
     if (!rstr && !result && num_bytes && buffer[0]>=0x20 && buffer[0]!=0x7f
         && !(xevent.state & (Mod1Mask|ControlMask))) {
       rstr = (char *)malloc(num_bytes + 1);
@@ -389,11 +451,14 @@ gtk_im_context_gcin_filter_keypress (GtkIMContext *context,
     result = gcin_im_client_forward_key_release(context_xim->gcin_ch,
       keysym, xevent.state, &rstr);
   }
-
-//  printf("event->type:%d iiiii %d\n", event->type, result);
-
+#if 0
+  printf("event->type:%d iiiii %d  %x %d rstr:%x\n", event->type, result, keysym,
+    num_bytes, rstr);
+#endif
   if (rstr) {
-//    printf("emit %s\n", rstr);
+#if DBG
+    printf("emit %s\n", rstr);
+#endif
     g_signal_emit_by_name (context, "commit", rstr);
     free(rstr);
   }
@@ -406,11 +471,12 @@ static void
 gtk_im_context_gcin_focus_in (GtkIMContext *context)
 {
   GtkIMContextGCIN *context_xim = GTK_IM_CONTEXT_GCIN (context);
-
-//  printf("gtk_im_context_gcin_focus_in\n");
+#if DBG
+  printf("gtk_im_context_gcin_focus_in\n");
+#endif
   if (context_xim->gcin_ch) {
 #if NEW_GTK_IM
-    if (context_xim->is_mozilla)
+    if (context_xim->is_mozilla && !context_xim->dirty_fix_off)
       add_cursor_timeout(context_xim);
 #endif
     gcin_im_client_focus_in(context_xim->gcin_ch);
@@ -440,7 +506,9 @@ static void
 gtk_im_context_gcin_set_cursor_location (GtkIMContext *context,
 					GdkRectangle *area)
 {
-//  printf("gtk_im_context_gcin_set_cursor_location %d\n", area->x);
+#if DBG
+  printf("gtk_im_context_gcin_set_cursor_location %d\n", area->x);
+#endif
   GtkIMContextGCIN *context_xim = GTK_IM_CONTEXT_GCIN (context);
 
   if (context_xim->gcin_ch) {
@@ -463,7 +531,10 @@ gtk_im_context_gcin_set_use_preedit (GtkIMContext *context,
 static void
 gtk_im_context_gcin_reset (GtkIMContext *context)
 {
-//  printf("gtk_im_context_gcin_reset\n");
+  GtkIMContextGCIN *context_gcin = GTK_IM_CONTEXT_GCIN (context);
+#if DBG
+  printf("gtk_im_context_gcin_reset %x\n", context_gcin);
+#endif
 }
 
 /* Mask of feedback bits that we render
@@ -476,7 +547,9 @@ gtk_im_context_gcin_get_preedit_string (GtkIMContext   *context,
 				       PangoAttrList **attrs,
                                        gint           *cursor_pos)
 {
-//  printf("gtk_im_context_gcin_get_preedit_string %x %x %x\n", str, attrs, cursor_pos);
+#if DBG
+  printf("gtk_im_context_gcin_get_preedit_string %x %x %x\n", str, attrs, cursor_pos);
+#endif
   GtkIMContextGCIN *context_xim = GTK_IM_CONTEXT_GCIN (context);
 
   if (str)
@@ -499,7 +572,7 @@ gtk_im_context_gcin_get_preedit_string (GtkIMContext   *context,
 void
 gtk_im_context_gcin_shutdown (void)
 {
-#if NEW_GTK_IM && 0
-  cancel_timeout();
+#if DBG
+ printf("shutdown\n");
 #endif
 }
