@@ -13,9 +13,12 @@
 #include "gcin-protocol.h"
 #include "gcin-im-client.h"
 
+static int flags_backup;
+
 static void save_old_sigaction_single(int signo, struct sigaction *act)
 {
   sigaction(signo, NULL, act);
+
   if (act->sa_handler != SIG_IGN)
     signal(signo, SIG_IGN);
 }
@@ -43,11 +46,10 @@ static GCIN_client_handle *gcin_im_client_reopen(GCIN_client_handle *gcin_ch, Di
 //  dbg("gcin_im_client_reopen\n");
 
   if (!dpy) {
-    dbg("null disp\n");
+    dbg("null disp %d\n", gcin_ch->fd);
     goto next;
   }
 
-
   Atom gcin_addr_atom = get_gcin_addr_atom(dpy);
   Window gcin_win = None;
 
@@ -56,7 +58,7 @@ static GCIN_client_handle *gcin_im_client_reopen(GCIN_client_handle *gcin_ch, Di
   int loop;
   for(loop=0; loop < MAX_TRY; loop++) {
     if ((gcin_addr_atom && (gcin_win=XGetSelectionOwner(dpy, gcin_addr_atom))!=None)
-        || getenv("GCIN_IM_CLEINT_NO_AUTO_EXEC"))
+        || getenv("GCIN_IM_CLIENT_NO_AUTO_EXEC"))
       break;
     static time_t exec_time;
 
@@ -71,7 +73,7 @@ static GCIN_client_handle *gcin_im_client_reopen(GCIN_client_handle *gcin_ch, Di
       save_old_sigaction_single(SIGCHLD, &ori_act);
 
       if ((pid=fork())==0) {
-#if 	FREEBSD
+#if     FREEBSD
         setpgid(0, getpid());
 #else
         setpgrp();
@@ -89,12 +91,43 @@ static GCIN_client_handle *gcin_im_client_reopen(GCIN_client_handle *gcin_ch, Di
     goto next;
   }
 
+  Atom actual_type;
+  int actual_format;
+  u_long nitems,bytes_after;
+  char *message_sock = NULL;
+  Atom gcin_sockpath_atom = get_gcin_sockpath_atom(dpy);
+
+//  printf("gcin_sockpath_atom %d\n", gcin_sockpath_atom);
+
+  if (!gcin_sockpath_atom || XGetWindowProperty(dpy, gcin_win, gcin_sockpath_atom, 0, 64,
+     False, AnyPropertyType, &actual_type, &actual_format,
+     &nitems,&bytes_after,(u_char **)&message_sock) != Success) {
+#if DBG || 1
+    dbg("XGetWindowProperty 2: old version of gcin or gcin is not running ??\n");
+#endif
+    goto next;
+  }
+
+  Server_sock_path srv_sock_path;
+  srv_sock_path.sock_path[0] = 0;
+  if (message_sock) {
+    memcpy(&srv_sock_path, message_sock, sizeof(srv_sock_path));
+    XFree(message_sock);
+  } else
+    goto next;
+
   struct sockaddr_un serv_addr;
   bzero((char *) &serv_addr,sizeof(serv_addr));
   serv_addr.sun_family = AF_UNIX;
   char sock_path[128];
 
-  get_gcin_im_srv_sock_path(sock_path, sizeof(sock_path));
+  if (srv_sock_path.sock_path[0]) {
+    strcpy(sock_path, srv_sock_path.sock_path);
+  }
+  else {
+    get_gcin_im_srv_sock_path(sock_path, sizeof(sock_path));
+  }
+
   addr = sock_path;
   strcpy(serv_addr.sun_path, sock_path);
 #ifdef SUN_LEN
@@ -119,12 +152,11 @@ static GCIN_client_handle *gcin_im_client_reopen(GCIN_client_handle *gcin_ch, Di
   goto next;
 
   struct sockaddr_in in_serv_addr;
-  Atom actual_type;
-  int actual_format;
-  u_long nitems,bytes_after;
-  char *message = NULL;
+  char *message;
 
 tcp:
+  message = NULL;
+
   if (!gcin_addr_atom || XGetWindowProperty(dpy, gcin_win, gcin_addr_atom, 0, 64,
      False, AnyPropertyType, &actual_type, &actual_format,
      &nitems,&bytes_after,(u_char **)&message) != Success) {
@@ -134,13 +166,13 @@ tcp:
     goto next;
   }
 
-
   if (message) {
     memcpy(&srv_ip_port, message, sizeof(srv_ip_port));
     XFree(message);
   } else
     goto next;
 
+
 //  dbg("im server tcp port %d\n", ntohs(srv_ip_port.port));
 
   bzero((char *) &in_serv_addr, sizeof(in_serv_addr));
@@ -196,8 +228,12 @@ next:
     }
   }
 
-  if (handle->fd && BITON(handle->flag, FLAG_GCIN_client_handle_has_focus))
-    gcin_im_client_focus_in(handle);
+  if (handle->fd)  {
+    if (BITON(handle->flag, FLAG_GCIN_client_handle_has_focus))
+      gcin_im_client_focus_in(handle);
+    int rstatus;
+    gcin_im_client_set_flags(handle, flags_backup, &rstatus);
+  }
 
   return handle;
 }
@@ -231,9 +267,11 @@ void gcin_im_client_close(GCIN_client_handle *handle)
 
 
 
-static void gen_req(GCIN_client_handle *handle, u_int req_no, GCIN_req *req)
+static int gen_req(GCIN_client_handle *handle, u_int req_no, GCIN_req *req)
 {
   validate_handle(handle);
+  if (!handle->fd)
+    return 0;
 
   bzero(req, sizeof(GCIN_req));
 
@@ -250,6 +288,8 @@ static void gen_req(GCIN_client_handle *handle, u_int req_no, GCIN_req *req)
   req->spot_location.y = handle->spot_location.y;
   to_gcin_endian_2(&req->spot_location.x);
   to_gcin_endian_2(&req->spot_location.y);
+
+  return 1;
 }
 
 static void error_proc(GCIN_client_handle *handle, char *msg)
@@ -264,7 +304,6 @@ static void error_proc(GCIN_client_handle *handle, char *msg)
 
 typedef struct {
   struct sigaction apipe;
-  struct sigaction aio;
 } SAVE_ACT;
 
 
@@ -273,14 +312,12 @@ typedef struct {
 static void save_old_sigaction(SAVE_ACT *save_act)
 {
   save_old_sigaction_single(SIGPIPE, &save_act->apipe);
-  save_old_sigaction_single(SIGIO, &save_act->aio);
 }
 
 
 static void restore_old_sigaction(SAVE_ACT *save_act)
 {
   restore_old_sigaction_single(SIGPIPE, &save_act->apipe);
-  restore_old_sigaction_single(SIGIO, &save_act->aio);
 }
 
 static int handle_read(GCIN_client_handle *handle, void *ptr, int n)
@@ -291,10 +328,13 @@ static int handle_read(GCIN_client_handle *handle, void *ptr, int n)
     return 0;
 
   SAVE_ACT save_act;
-
+#if 1
   save_old_sigaction(&save_act);
+#endif
   int r = read(fd, ptr, n);
+#if 1
   restore_old_sigaction(&save_act);
+#endif
 
   if (r<=0)
     return r;
@@ -321,11 +361,13 @@ static int handle_write(GCIN_client_handle *handle, void *ptr, int n)
 
 
   SAVE_ACT save_act;
-
+#if 1
   save_old_sigaction(&save_act);
+#endif
   int r =  write(fd, tmp, n);
+#if 1
   restore_old_sigaction(&save_act);
-
+#endif
   free(tmp);
 
   return r;
@@ -338,7 +380,8 @@ void gcin_im_client_focus_in(GCIN_client_handle *handle)
   handle->flag |= FLAG_GCIN_client_handle_has_focus;
 
   GCIN_req req;
-  gen_req(handle, GCIN_req_focus_in, &req);
+  if (!gen_req(handle, GCIN_req_focus_in, &req))
+    return;
 
   if (handle_write(handle, &req, sizeof(req)) <=0) {
     error_proc(handle,"gcin_im_client_focus_in error");
@@ -354,7 +397,8 @@ void gcin_im_client_focus_out(GCIN_client_handle *handle)
   handle->flag &= ~FLAG_GCIN_client_handle_has_focus;
 
   GCIN_req req;
-  gen_req(handle, GCIN_req_focus_out, &req);
+  if (!gen_req(handle, GCIN_req_focus_out, &req))
+    return;
 
   if (handle_write(handle, &req, sizeof(req)) <=0) {
     error_proc(handle,"gcin_im_client_focus_out error");
@@ -369,13 +413,10 @@ static int gcin_im_client_forward_key_event(GCIN_client_handle *handle,
 {
   *rstr = NULL;
 
-  if (!handle->fd) {
-//    dbg("gcin_im_client_forward_key_event fd is 0\n");
-    return 0;
-  }
-
   GCIN_req req;
-  gen_req(handle, event_type, &req);
+  if (!gen_req(handle, event_type, &req))
+    return 0;
+
   req.keyeve.key = key;
   to_gcin_endian_4(&req.keyeve.key);
   req.keyeve.state = state;
@@ -394,7 +435,6 @@ static int gcin_im_client_forward_key_event(GCIN_client_handle *handle,
     return FALSE;
   }
 
-
   to_gcin_endian_4(&reply.datalen);
   to_gcin_endian_4(&reply.flag);
 
@@ -409,7 +449,7 @@ static int gcin_im_client_forward_key_event(GCIN_client_handle *handle,
 
 //  dbg("gcin_im_client_forward_key_event %x\n", reply.flag);
 
-  return ((reply.flag & GCIN_reply_key_processed) !=0 );
+  return reply.flag;
 }
 
 
@@ -418,6 +458,8 @@ int gcin_im_client_forward_key_press(GCIN_client_handle *handle,
                                           KeySym key, u_int state,
                                           char **rstr)
 {
+  if (!handle)
+    return 0;
   // in case client didn't send focus in event
   if (!BITON(handle->flag, FLAG_GCIN_client_handle_has_focus)) {
     gcin_im_client_focus_in(handle);
@@ -427,8 +469,10 @@ int gcin_im_client_forward_key_press(GCIN_client_handle *handle,
   }
 
 //  dbg("gcin_im_client_forward_key_press\n");
-  return gcin_im_client_forward_key_event(
+ int flag = gcin_im_client_forward_key_event(
              handle, GCIN_req_key_press, key, state, rstr);
+
+  return ((flag & GCIN_reply_key_processed) !=0);
 }
 
 
@@ -437,10 +481,13 @@ int gcin_im_client_forward_key_release(GCIN_client_handle *handle,
                                           KeySym key, u_int state,
                                           char **rstr)
 {
+  if (!handle)
+    return 0;
   handle->flag |= FLAG_GCIN_client_handle_has_focus;
 //  dbg("gcin_im_client_forward_key_release\n");
-  return gcin_im_client_forward_key_event(
+  int flag = gcin_im_client_forward_key_event(
              handle, GCIN_req_key_release, key, state, rstr);
+  return ((flag & GCIN_reply_key_processed) !=0);
 }
 
 
@@ -453,7 +500,8 @@ void gcin_im_client_set_cursor_location(GCIN_client_handle *handle, int x, int y
     return;
 
   GCIN_req req;
-  gen_req(handle, GCIN_req_set_cursor_location, &req);
+  if (!gen_req(handle, GCIN_req_set_cursor_location, &req))
+    return;
 
   if (handle_write(handle, &req, sizeof(req)) <=0) {
     error_proc(handle,"gcin_im_client_set_cursor_location error");
@@ -471,14 +519,22 @@ void gcin_im_client_set_window(GCIN_client_handle *handle, Window win)
 }
 
 
-void gcin_im_client_set_flags(GCIN_client_handle *handle, int flags)
+void gcin_im_client_set_flags(GCIN_client_handle *handle, int flags, int *ret_flag)
 {
   GCIN_req req;
 
-  gen_req(handle, GCIN_req_set_flags, &req);
+  if (!gen_req(handle, GCIN_req_set_flags, &req))
+    return;
+
   req.flag |= flags;
 
+  flags_backup = flags;
+
   if (handle_write(handle, &req, sizeof(req)) <=0) {
     error_proc(handle,"gcin_im_client_set_flags error");
   }
+
+  if (handle_read(handle, ret_flag, sizeof(int)) <= 0) {
+    error_proc(handle, "cannot read reply str from gcin server");
+  }
 }
