@@ -10,9 +10,6 @@ typedef enum {
   GCIN_req_set_flags = 0x20
 } GCIN_req_t;
 
-enum {
-  GCIN_flag_has_key_release = 1,
-};
 
 typedef struct {
 #if 0
@@ -48,6 +45,7 @@ typedef struct {
 
 enum {
   GCIN_reply_key_processed = 1,
+  GCIN_reply_key_state_disabled = 2,
 };
 
 
@@ -71,6 +69,9 @@ typedef struct {
   GCIN_PASSWD passwd;
 } Server_IP_port;
 
+typedef struct {
+  char sock_path[80];
+} Server_sock_path;
 
 void __gcin_enc_mem(u_char *p, int n, GCIN_PASSWD *passwd, u_int *seed);
 
