@@ -7,6 +7,7 @@
 #include "pho.h"
 #include "tsin.h"
 #include "gcin-conf.h"
+#include "tsin-parse.h"
 
 static gint64 key_press_time;
 extern gboolean b_hsu_kbm;
@@ -35,7 +36,7 @@ static int save_frm, save_to;
 static int current_page;
 static int startf;
 static gboolean full_match;
-static gboolean tsin_half_full;
+gboolean tsin_half_full;
 static gboolean tsin_buffer_editing = 0;
 
 typedef struct {
@@ -63,7 +64,7 @@ gboolean pho_has_input();
 
 gboolean tsin_has_input()
 {
-  return c_len || pho_has_input() || !eng_ph;
+  return c_len || pho_has_input();
 }
 
 
@@ -513,7 +514,6 @@ static void put_b5_char(char *b5ch, phokey_t key)
    init_chpho_i(c_idx);
    bchcpy(chpho[c_idx].ch, b5ch);
    bchcpy(chpho[c_idx].och, b5ch);
-   bchcpy(chpho[c_idx].ph1ch, b5ch);
 
    disp_char_chbuf(c_idx);
 
@@ -545,14 +545,12 @@ static gboolean chpho_eq_pho(int idx, phokey_t *phos, int len)
 }
 
 
-static void get_sel_phrase()
+static void get_sel_phrase0(int selidx, gboolean eqlen)
 {
   int sti,edi,j;
   u_char len, mlen;
 
-  phrase_count = 0;
-
-  mlen=c_len-c_idx;
+  mlen=c_len-selidx;
 
   if (!mlen)
     return;
@@ -561,7 +559,7 @@ static void get_sel_phrase()
     mlen=MAX_PHRASE_LEN;
 
   phokey_t pp[MAX_PHRASE_LEN + 1];
-  extract_pho(c_idx, mlen, pp);
+  extract_pho(selidx, mlen, pp);
 
   if (!tsin_seek(pp, 2, &sti, &edi))
     return;
@@ -573,12 +571,12 @@ static void get_sel_phrase()
 
     load_tsin_entry(sti, &len, &usecount, stk, stch);
 
-    if (len > mlen || len==1) {
+    if (eqlen && len!=mlen || (!eqlen && len > mlen) || len==1) {
       sti++;
       continue;
     }
 
-    if (chpho_eq_pho(c_idx, stk, len)) {
+    if (chpho_eq_pho(selidx, stk, len)) {
       sellen[phrase_count]=len;
       utf8cpyN(selstr[phrase_count++], stch, len);
     }
@@ -587,6 +585,25 @@ static void get_sel_phrase()
   }
 }
 
+static void get_sel_phrase_end()
+{
+  int stidx = c_idx - 5;
+  if (stidx < 0)
+    stidx = 0;
+
+  phrase_count = 0;
+  int i;
+  for(i=stidx; i < c_len - 1; i++) {
+    get_sel_phrase0(i, TRUE);
+  }
+}
+
+static void get_sel_phrase()
+{
+  phrase_count = 0;
+  get_sel_phrase0(c_idx, FALSE);
+}
+
 static void get_sel_pho()
 {
   phokey_t key;
@@ -621,7 +638,6 @@ static void get_sel_pho()
 
 void clear_sele();
 void set_sele_text(int i, char *text, int len);
-void disp_selections(int idx);
 void disp_arrow_up(), disp_arrow_down();
 
 static void disp_current_sel_page()
@@ -651,7 +667,7 @@ static void disp_current_sel_page()
   if (current_page > 0)
     disp_arrow_up();
 
-  disp_selections(c_idx==c_len?c_idx-1:c_idx);
+  disp_tsin_select(c_idx==c_len?c_idx-1:c_idx);
 }
 
 static int fetch_user_selection(int val, char **seltext)
@@ -856,7 +872,7 @@ static void disp_pre_sel_page()
 #if 0
   dbg("ph_sta:%d\n", ph_sta);
 #endif
-  disp_selections(ph_sta);
+  disp_tsin_select(ph_sta);
 }
 
 static void close_selection_win()
@@ -874,6 +890,9 @@ void tsin_set_eng_ch(int nmod)
   show_stat();
   drawcursor();
 
+  if (!eng_ph)
+    clrin_pho_tsin();
+
   show_button_pho(eng_ph);
 #if TRAY_ENABLED
   load_tray_icon();
@@ -894,6 +913,7 @@ void tsin_toggle_half_full()
     tsin_half_full^=1;
     key_press_time = 0;
     drawcursor();
+    load_tray_icon();
 }
 
 
@@ -1162,7 +1182,12 @@ static gboolean pre_punctuation(KeySym xkey)
     int c = p - shift_punc;
     phokey_t key=0;
 
-    return add_to_tsin_buf(chars[c], &key, 1);
+    if (c_len)
+      return add_to_tsin_buf(chars[c], &key, 1);
+    else {
+      send_text(chars[c]);
+      return 1;
+    }
   }
 
   return 0;
@@ -1193,9 +1218,7 @@ gint64 current_time();
 
 static void call_tsin_parse()
 {
-  TSIN_PARSE parse[MAX_PH_BF_EXT+1];
-  bzero(parse, sizeof(parse));
-  tsin_parse(parse);
+  tsin_parse();
   prbuf();
 }
 
@@ -1330,8 +1353,9 @@ int feedkey_pp(KeySym xkey, int kbstate)
   int shift_m=kbstate&ShiftMask;
   int j,jj,kk, idx;
   char kno;
+  int caps_eng_tog = tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_CapsLock;
 
-  if (tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_CapsLock)
+  if (caps_eng_tog)
     eng_ph = !(kbstate&LockMask);
 
   key_press_time = 0;
@@ -1345,7 +1369,22 @@ int feedkey_pp(KeySym xkey, int kbstate)
 
    if (!eng_ph && !c_len && gcin_pop_up_win && xkey!=XK_Caps_Lock) {
      hide_win0();
-     return 0;
+
+     gboolean is_ascii = xkey>=' ' && xkey<0x7f;
+
+     if (caps_eng_tog && is_ascii) {
+       case_inverse(&xkey, shift_m);
+       send_ascii(xkey);
+       return 1;
+     }
+     else {
+       if (tsin_half_full && is_ascii) {
+         send_text(half_char_to_full_char(xkey));
+         return 1;
+       }
+       else
+         return 0;
+     }
    }
 
    int o_sel_pho = sel_pho;
@@ -1401,7 +1440,7 @@ int feedkey_pp(KeySym xkey, int kbstate)
      case XK_Right:
         return cursor_right();
      case XK_Caps_Lock:
-        if (tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_CapsLock) {
+        if (caps_eng_tog) {
           close_selection_win();
           tsin_toggle_eng_ch();
           return 1;
@@ -1447,11 +1486,13 @@ tab_phrase_end:
               disp_pre_sel_page();
             }
 
-            if (!c_len && gcin_pop_up_win && !j)
+            if (!c_len && gcin_pop_up_win && typ_pho_empty())
               hide_win0();
+
             return 1;
           }
 
+
         if (!c_idx)
           return 0;
 
@@ -1531,10 +1572,10 @@ tab_phrase_end:
          close_selection_win();
          goto asc_char;
        }
-     case XK_Down:
-       if (!eng_ph && xkey == XK_space)
-           goto asc_char;
 
+       if (!eng_ph)
+           goto asc_char;
+     case XK_Down:
        if (!ityp3_pho && (typ_pho[0]||typ_pho[1]||typ_pho[2]) && xkey==XK_space) {
          ctyp=3;
          kno=0;
@@ -1552,7 +1593,11 @@ change_char:
          return 1;
 
        if (!sel_pho) {
-         get_sel_phrase();
+         if (c_idx==c_len) {
+           get_sel_phrase_end();
+         } else
+           get_sel_phrase();
+
          get_sel_pho();
          sel_pho=1;
          current_page = 0;
@@ -1609,46 +1654,29 @@ other_keys:
            return cursor_delete();
        }
 
-       if (tsin_space_opt == TSIN_SPACE_OPT_INPUT)
-         xkey_lcase = keypad_proc(xkey);
-
        char *pp;
        if ((pp=strchr(phkbm.selkey,xkey_lcase)) && sel_pho) {
          int c=pp-phkbm.selkey;
          char *sel_text;
          int len = fetch_user_selection(c, &sel_text);
+         int cpsta = chpho[c_idx].psta;
+         int sel_idx = c_idx;
+         if (c_idx == c_len)
+           sel_idx = c_len - len;
 
-         if (len > 1) {
-           int cpsta = chpho[c_idx].psta;
-           if (cpsta >= 0)
-             set_chpho_ch(&chpho[c_idx], sel_text, len);
-           else
-             set_chpho_ch2(&chpho[c_idx], sel_text, len);
+         if (cpsta >= 0)
+           set_chpho_ch(&chpho[sel_idx], sel_text, len);
+         else
+           set_chpho_ch2(&chpho[sel_idx], sel_text, len);
 
-           chpho[c_idx].flag &= ~FLAG_CHPHO_PHRASE_VOID;
-           set_fixed(c_idx, len);
+//         chpho[c_idx].flag &= ~FLAG_CHPHO_PHRASE_VOID;
+         set_fixed(sel_idx, len);
 
-           call_tsin_parse();
+         call_tsin_parse();
 
-           if (c_idx + len == c_len) {
-             ph_sta = -1;
-             draw_ul(c_idx, c_len);
-           }
-         } else
-         if (len == 1) { // single chinese char
-           i= c_idx==c_len?c_idx-1:c_idx;
-           key=chpho[i].pho;
-           set_chpho_ch(&chpho[i], sel_text, 1);
-
-           if (i && chpho[i].psta == i-1 && !(chpho[i-1].flag & FLAG_CHPHO_FIXED)) {
-             set_chpho_ch(&chpho[i-1], chpho[i-1].ph1ch, 1);
-#if 0
-             chpho[i-1].flag |= FLAG_CHPHO_PHRASE_VOID;
-#endif
-           }
-
-           set_fixed(i, 1);
-           call_tsin_parse();
+         if (c_idx + len == c_len) {
+           ph_sta = -1;
+           draw_ul(c_idx, c_len);
          }
 
          if (len) {
@@ -1675,7 +1703,15 @@ other_keys:
    if (xkey > 0x7e && !key_pad)
      return 0;
 
+   if (key_pad && !c_len)
+     return 0;
+
    if (!eng_ph || typ_pho[0]!=BACK_QUOTE_NO && (shift_m || key_pad || !phkbm.phokbm[xkey][0].num)) {
+       if (eng_ph && !shift_m && strchr(hsu_punc, xkey) && !phkbm.phokbm[xkey][0].num) {
+         if (pre_punctuation_hsu(xkey))
+           return 1;
+       }
+
        if (key_pad)
          xkey = key_pad;
 asc_char:
@@ -1689,22 +1725,12 @@ asc_char:
             return 1;
         }
 
-
-//        dbg("xkey: %c\n", xkey);
-
         if (shift_m && eng_ph)  {
           char *ppp=strchr(ochars,xkey);
 
           if (!(kbstate&LockMask) && ppp && !((ppp-ochars) & 1))
             xkey=*(ppp+1);
 
-#if 0
-          if (kbstate&LockMask && islower(xkey))
-            xkey-=0x20;
-          else
-            if (!(kbstate&LockMask) && isupper(xkey))
-              xkey+=0x20;
-#endif
         } else {
           if (!eng_ph && tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_CapsLock
               && gcin_capslock_lower) {
@@ -1714,17 +1740,26 @@ asc_char:
 
         if (xkey > 127)
           return 0;
+        char tstr[CH_SZ + 1];
+        bzero(tstr, sizeof(tstr));
 
         u_char tt=xkey;
-        shift_ins();
 
         if (tsin_half_full) {
-          bchcpy(chpho[c_idx].ch, half_char_to_full_char(xkey));
+          strcpy(tstr, half_char_to_full_char(xkey));
         } else {
-//          dbg("%c\n", tt);
-          chpho[c_idx].ch[0]=tt;
+          tstr[0] = tt;
         }
 
+        if (!c_len) {
+          send_text(tstr);
+          return 1;
+        }
+
+        shift_ins();
+
+        memcpy(chpho[c_idx].ch, tstr, CH_SZ);
+
         set_fixed(c_idx, 1);
         phokey_t tphokeys[32];
         tphokeys[0]=0;
@@ -1746,9 +1781,8 @@ asc_char:
        return 0;
      }
 
+
      // for hsu & et26
-     if (strchr(hsu_punc, xkey) && !phkbm.phokbm[xkey][0].num && typ_pho[0]!=BACK_QUOTE_NO)
-       return pre_punctuation_hsu(xkey);
 
      if (xkey >= 'A' && xkey <='Z' && typ_pho[0]!=BACK_QUOTE_NO)
        xkey+=0x20;
@@ -1836,6 +1870,7 @@ llll2:
 
      ii=idx_pho[vv].start;
      start_idx=ii;
+     stop_idx = idx_pho[vv+1].start;
 #if 0
      printf("%x %x %d vv:%d idxnum_pho:%d-->", ttt, key, start_idx, vv, idxnum_pho);
      utf8_putchar(ch_pho[start_idx].ch);
@@ -1843,7 +1878,10 @@ llll2:
 #endif
    } /* pho */
 
-   put_b5_char(ch_pho[start_idx].ch, key);
+   if (!c_len && typ_pho[0]==BACK_QUOTE_NO && stop_idx - start_idx == 1)
+     send_text(ch_pho[start_idx].ch);  // it's ok since ,. are 3 byte
+   else
+     put_b5_char(ch_pho[start_idx].ch, key);
 
    call_tsin_parse();
 
@@ -1853,7 +1891,6 @@ llll2:
    drawcursor();
    hide_pre_sel();
 
-
    if (ph_sta < 0) {
 restart:
      if ((match_len=scanphr(c_idx-1,1, FALSE)))
@@ -1946,11 +1983,6 @@ restart:
 
        if (j < mdist)
          continue;
-#if 0
-       ch_pho_cpy(&chpho[ph_sta], pre_sel[i].str, pre_sel[i].phokey, mdist);
-       if (chpho[ph_sta].psta < 0)
-         set_chpho_ch2(&chpho[ph_sta], pre_sel[i].str, mdist);
-#endif
 
        int j;
        for(j=0;j < mdist; j++) {
@@ -1970,7 +2002,6 @@ restart:
          if (pre_selN == 1)
            pre_selN = 0;
        }
-
        return 1;
      }
    }
